@@ -24,7 +24,7 @@
 # Python modules
 import os, sys, time, signal
 import urllib, re, optparse
-import inspect, threading
+import inspect
 
 # Xlib modules
 import Xlib.display, Xlib.X, Xlib.XK
@@ -36,9 +36,9 @@ pygtk.require("2.0")
 import gtk, gtk.glade, gobject
 
 ####################################
-__version__ = "$Revision: 1.9 $"
+__version__ = "0.9.2"
 __author__ = "Arnau Sanchez <arnau@ehas.org>"
-__depends__ = ['Xlib', 'PyGTK2', 'Glade', 'Python-2.3']
+__depends__ = ['Xlib', 'PyGTK2', 'PyGlade2', 'Python-2.3']
 __copyright__ = """Copyright (C) 2006 Arnau Sanchez <arnau@ehas.org>.
 This code is distributed under the terms of the GNU General Public License."""
 
@@ -47,25 +47,26 @@ NAME = "xhotkeys"
 DISABLE_HOTKEY = Xlib.XK.XK_BackSpace
 DISABLED_STRING = "Disabled"
 MOUSE_STRING = "Button"
+TARGET_TYPE_URI_LIST = 80
+TABLE = ("Name", 100, False), ("Command", 300, True), ("Hotkey", 150, True)
 SETTINGS_WINDOW = "window: %d,%d (%dx%d)"
 SETTINGS_COLUMN = "table: %d,%d,%d"
-TARGET_TYPE_URI_LIST = 80
+SETTINGS_ORDER = "columns order: %s,%d"
 
-## Verbose levels
+### Verbose levels
 LEVELS_STRING = "error", "info", "debug"
 ERROR, INFO, DEBUG = range(3)
 
-x = Xlib.X
-MODIFIERS = [("shift", x.ShiftMask), ("lock", x.LockMask), ("control", x.ControlMask), \
-	("alt", x.Mod1Mask), ("mod2", x.Mod2Mask), ("mod3", x.Mod3Mask), \
-	("winkey", x.Mod4Mask), ("altgr", x.Mod5Mask)]
+### Modifiers ordered by ModifierMask
+## Names changed for clarity: mod1->alt, mod4->winkey, mod5->altgr 
+MODIFIERS = "shift", "lock", "control", "alt", "mod2", "mod3", "winkey", "altgr"
 
 ########################################
 ## From Peter Norvig's Infrequently Answered Questions
 def create_dict(**dict): 
 	return dict
 
-#######################################################
+####################################################
 def debug(text, verbose_level, level, exit=None):
 	"""Give some debug messages.
 	
@@ -88,7 +89,7 @@ def run_command(command):
 
 ######################################
 ######################################
-class XhotkeysServer(threading.Thread):
+class XhotkeysServer:
 	"""Wait for key combinations and run command if configured"""
 
 	##################################################
@@ -100,15 +101,14 @@ class XhotkeysServer(threading.Thread):
 		self.combinations = []
 		self.init_signals([signal.SIGCHLD])
 		self.get_keylock_masks()
-		self.stop_flag = False
-		threading.Thread.__init__(self)
 
 	########################################
 	def string_to_mask(self, strmods):
 		"""Return modifier mask from its name: "shift", "alt", ..."""
 		mask = 0
 		for strmod in strmods:
-			mask |= dict(MODIFIERS)[strmod.lower()]
+			try: mask |= 1 << list(MODIFIERS).index(strmod.lower())
+			except: self.debug("modifier not available: %s" %strmod, ERROR)
 		return mask
 
 	########################################
@@ -118,19 +118,22 @@ class XhotkeysServer(threading.Thread):
 		if strcode.find(MOUSE_STRING) == 0:
 			return "button", int(strcode[-1])
 		for key in inspect.getmembers(Xlib.XK):
-			if key[0].find("XK_") != 0 or strcode.lower() != key[0].replace("XK_", "").lower(): continue
+			if key[0].find("XK_") != 0 or strcode.lower() != key[0].replace("XK_", "").lower(): 
+				continue
 			return "key", self.display.keysym_to_keycode(key[1])
 
 	##################################################
 	def get_keylock_masks(self):
-		"""Get Caps/Num/Scroll Lock masks. Needed because must their value 
+		"""Get Caps/Num/Scroll Lock masks. Needed because their value 
 		must be ignored in key combinations"""
 		kc, ktk = Xlib.XK, self.display.keysym_to_keycode
 		keycodes = {ktk(kc.XK_Caps_Lock): "caps", ktk(kc.XK_Num_Lock): "num", ktk(kc.XK_Scroll_Lock): "scroll"}
 		self.locks_mask = dict([(x, 0) for x in keycodes.values()])
+		self.modifiers_mask = 2**len(MODIFIERS)-1
 		for index, mask in enumerate(self.display.get_modifier_mapping()):
 			try: self.locks_mask[keycodes[mask[0]]] = 1 << index
 			except: pass
+			else: self.modifiers_mask &= ~(1 << index)
 
 	##################################################
 	def init_signals(self, signals):
@@ -161,7 +164,7 @@ class XhotkeysServer(threading.Thread):
 		debug(text, self.verbose_level, level, exit)
 
 	#######################################
-	def ungrab(self):
+	def stop(self):
 		"""Ungrab keyboard and mouse bindings"""
 		self.display.ungrab_keyboard(Xlib.X.CurrentTime)
 		self.display.flush()
@@ -171,38 +174,32 @@ class XhotkeysServer(threading.Thread):
 	#######################################
 	def set_combination(self, combinations):
 		"""Set the current hotkey table"""
-		self.ungrab()
+		self.stop()
 		self.combinations = []
 		for comb in combinations:
 			hotkey = comb["hotkey"]
-			type = mask = keycode = 0
+			hktype = mask = keycode = 0
 			if hotkey.lower() != DISABLED_STRING.lower():
 				modifiers = re.findall("<(\w+)>", hotkey)
 				mask = self.string_to_mask(modifiers)
-				try: key = re.findall("(@?\w+@?)$", hotkey)[0]
-				except: self.debug("error parsing hotkey: %s" %hotkey, ERROR); continue
-				try: type, keycode = self.string_to_keycode(key)
-				except: self.debug("error parsing: %s" %key, ERROR); continue
+				try: 
+					key = re.findall("(@?\w+@?)$", hotkey)[0]
+					hktype, keycode = self.string_to_keycode(key)
+				except: 
+					self.debug("error parsing hotkey: %s (entry %s ignored)" %(key, comb["name"]), ERROR); continue
 			newcomb = create_dict(name=comb["name"], command=comb["command"], \
-				type=type, mask=mask, keycode=keycode)
+				hktype=hktype, mask=mask, keycode=keycode)
 			self.combinations.append(newcomb)
-
 		self.configure()
 
-	##################################################
-	def stop(self):
-		"""Stop the thread"""
-		self.stop_flag = True
-
 	##################################
 	def grab_button(self, button, modifiers, owner_events, event_mask, \
 		pointer_mode, keyboard_mode, confine_to, cursor, onerror=None):
-		"""This need an explanation: at this moment, June 2006, Python-Xlib is broken
-		at grab_button(), a basic function to grab mouse events. A bug on this issue was 
-		already reported at the ends of 2004, with no response... the project seems 
-		temporally dead. 
+		"""At this moment, June 2006, Python-Xlib has a bug on the grab_button() 
+		function. A bug on this issue was already reported on the ends of 2004, with 
+		no response... the project seems temporally dead. 
 		
-		So, here we make a direct call to GrabButton() with the correct parameters. 
+		Here we make a direct call to GrabButton() with the correct parameters. 
 		The problem was that the "cursor=cursor" parameter was missing, that's all"""
 		Xlib.protocol.request.GrabButton(display=self.root.display, onerror=onerror, \
 			owner_events=owner_events, grab_window=self.root.id, event_mask=event_mask, \
@@ -218,37 +215,28 @@ class XhotkeysServer(threading.Thread):
 		self.root.change_attributes(event_mask=Xlib.X.KeyPressMask | Xlib.X.ButtonPressMask)
 		for comb in self.combinations:
 			if not comb["keycode"]: continue
-			key, mod, type = [comb[x] for x in ("keycode", "mask", "type")]
+			key, mod, hktype = [comb[x] for x in ("keycode", "mask", "hktype")]
 			cl, nl, sl = [self.locks_mask[x] for x in ("caps", "num", "scroll")]
 			for mask in (0, cl, nl, sl, cl | nl, cl | sl, cl | nl | sl):
-				if type == "button":
+				if hktype == "button":
 					self.grab_button(key, mod|mask, self.root, Xlib.X.ButtonPressMask, mode, mode, 0, 0)
-				elif type == "key":
+				elif hktype == "key":
 					self.root.grab_key(key, mod|mask, 0, mode, mode)
 		
 	##################################################
 	def run(self):
-		"""Main run thread. Read X events and look for configured hotkeys"""
-		while not self.stop_flag:
-			while self.display.pending_events():
-				event = self.display.next_event()
-				mask = event.state
-				self.debug(event)
-				if event.type == Xlib.X.KeyPress:
-					keycode = event.detail					
-					for comb in self.combinations:
-						if comb["type"] == "key" and comb["keycode"] == keycode and comb["mask"] == mask:
-							run_command(comb["command"])
-				elif event.type == Xlib.X.ButtonPress:
-					button = event.detail
-					for comb in self.combinations:
-						if comb["type"] == "button" and comb["keycode"] == button and comb["mask"] == mask:
-							run_command(comb["command"])
-			
-			# Give the CPU a breath (1/4 second should be fine)
-			time.sleep(0.25)
-		
-		self.ungrab()
+		"""Main method. Read X events and look for configured hotkeys"""
+		while self.display.pending_events():
+			event = self.display.next_event()
+			mask = event.state & self.modifiers_mask
+			self.debug(event)
+			event_to_type = {Xlib.X.KeyPress: "key", Xlib.X.ButtonPress: "button"}
+			if event.type not in (Xlib.X.KeyPress, Xlib.X.ButtonPress): continue
+			for command in [comb["command"] for comb in self.combinations if \
+				comb["hktype"] == event_to_type[event.type] and \
+				comb["keycode"] == event.detail and comb["mask"] == mask]:
+				run_command(command)
+	
 
 #######################
 #######################
@@ -262,16 +250,21 @@ class Xhotkeys:
 		self.display = Xlib.display.Display()
 		self.root = self.display.screen().root
 		self.hkserver = None
-		self.create_keysym_mask(MODIFIERS)
+		self.create_keysym_to_mask(MODIFIERS)
 
 	####################################
-	def create_keysym_mask(self, modifiers):
+	def create_keysym_to_mask(self, modifiers):
 		self.keysym_to_mask = {}
 		self.keysym_to_maskname = {}
+		xk = Xlib.XK
+		lock_keysyms = [xk.XK_Caps_Lock, xk.XK_Num_Lock, xk.XK_Scroll_Lock]
 		for maskindex, mmlist in enumerate(self.display.get_modifier_mapping()):
-			for ks in [self.display.keycode_to_keysym(mm, index) for mm in mmlist if mm != 0 for index in range(0,4)]:
-				if ks != Xlib.X.NoSymbol: 
-					self.keysym_to_maskname[ks], self.keysym_to_mask[ks] = modifiers[maskindex]
+			keysyms = [self.display.keycode_to_keysym(mm, index) for mm in mmlist if mm != 0 for index in range(4)]
+			for ks in keysyms:
+				if ks in lock_keysyms: keysyms = []; break
+			for ks in keysyms:
+				if ks != Xlib.X.NoSymbol and modifiers[maskindex]: 
+					self.keysym_to_maskname[ks], self.keysym_to_mask[ks] = modifiers[maskindex], 1<<maskindex
 		self.debug("keysym_to_maskname table: %s" %str(self.keysym_to_maskname))
 
 	####################################
@@ -285,7 +278,7 @@ class Xhotkeys:
 		self.debug("signal received: %s" %signame)
 		
 		if signum == signal.SIGHUP:
-			self.debug("send reload configuration command")
+			self.debug("reload configuration")
 			self.reload()
 		elif signum in (signal.SIGTERM, signal.SIGINT):
 			self.debug("%s stopped (pid %d)" %(NAME, os.getpid()), INFO)
@@ -312,7 +305,7 @@ class Xhotkeys:
 		mask = 0
 		for keycode in keycodes:
 			for keysym in self.keysym_to_mask:
-				if keycode in self.display.keysym_to_keycodes(keysym):
+				if keycode in [x[0] for x in self.display.keysym_to_keycodes(keysym)]:
 					mask |= self.keysym_to_mask[keysym]
 		return mask
 		
@@ -323,12 +316,7 @@ class Xhotkeys:
 		for keycode in keycodes:
 			for keysym in self.keysym_to_mask:
 				if keycode in [x[0] for x in self.display.keysym_to_keycodes(keysym)]:
-					string = self.keysym_to_maskname[keysym]
-					if string.lower() == "altgr": name = "AltGr"
-					elif string.lower() == "winkey": name = "WinKey"
-					else: name = string.title()
-					print name
-					keycodes_to_modifier[keycode] = name
+					keycodes_to_modifier[keycode] = self.keysym_to_maskname[keysym].title()
 		strmod = strkey = ""
 		current_modifiers = []
 		for keycode in keycodes:
@@ -354,7 +342,7 @@ class Xhotkeys:
 		try: fd = open(self.cfile)
 		except IOError: 
 			if not exit_on_error: return []
-			self.debug("configuration file not found: %s" %self.cfile)
+			self.debug("configuration file not found: %s" %self.cfile, INFO)
 			return self.combinations
 			
 		for numline, line in enumerate(fd.readlines()):
@@ -385,7 +373,7 @@ class Xhotkeys:
 
 	#######################################
 	def write_configuration(self):
-		self.debug("opening configuration file to get old configuration: %s" %self.cfile, INFO)
+		self.debug("opening configuration file to get old configuration: %s" %self.cfile, DEBUG)
 		try: fd = open(self.cfile, "r")
 		except IOError: original = []
 		else: original = fd.readlines()	; fd.close()
@@ -397,42 +385,33 @@ class Xhotkeys:
 			sys.exit(1)
 
 		entries = dict([ (comb["name"], comb["hotkey"] + ":" + comb["command"]) for comb in self.combinations])
-		entries_processed = []
+		combinations = self.combinations[:]
 		for oline in original:
+			if not combinations: break
 			line = oline.strip()
 			if not line: fd.write(oline); continue
 			try: name, value = line.split("=", 1)
 			except: fd.write(oline); continue
 			name, value = name.strip(), value.strip()
-			if name not in entries:
-				self.debug("old entry removed: %s" %oline.strip())
-				continue
-			if value == entries[name]:
+			comb_name = combinations[0]["name"]
+			if name != comb_name: self.debug("entry removed: %s" %oline.strip()); continue
+			if value == entries[comb_name]:
 				self.debug("keep old entry: %s" %oline.strip())
-				entries_processed.append(name)
 				fd.write(oline)
-				continue
 			else:
 				line = name + "=" + entries[name]
 				self.debug("modified entry saved: %s" %line)
-				entries_processed.append(name)
 				fd.write(line + "\n")
-				continue
+			del combinations[0]
 		
-		for entry, value in entries.items():
-			if entry in entries_processed: continue
-			line = entry + "=" + value
+		for comb in combinations:
+			name = comb["name"]
+			line = name + "=" + entries[name]
 			self.debug("new entry saved: %s" %line)
 			fd.write(line + "\n")
-				
+		self.debug("%s combinations written" %len(self.combinations), INFO)
 		fd.close()
 
-	#######################################
-	def end_combinations_keys(self):
-		if self.hkserver.isAlive():
-			self.hkserver.join(0.1)
-			gobject.idle_add(self.end_combinations_keys)
-
 	#######################################
 	def filechooser_file(self, widget):
 		command = widget.get_filename()
@@ -441,12 +420,11 @@ class Xhotkeys:
 
 	#######################################
 	def set_hotkey_text(self, text):
-		self.widgets["hotkey"].set_text(text)
+		self.widgets["hotkey"].set_text(text.title())
 		
 	#######################################
 	def error_message_response(self, widget, retval):
 		widget.destroy()
-		self.apply_widgets("set_sensitive", addwindow=True, hotkey=False, change=True)
 		self.set_hotkey_text(self.current_comb["hotkey"])
 
 	######################################
@@ -454,9 +432,7 @@ class Xhotkeys:
 		message = gtk.MessageDialog(None, gtk.DIALOG_MODAL, \
 			gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, text)
 		message.connect("response", self.error_message_response)
-		self.widgets["addwindow"].set_sensitive(False)
-		message.show()
-		return
+		message.run()
 
 	######################################
 	def hotkey_end(self, keycodes=None, hotkey=None, disable=False):
@@ -492,9 +468,13 @@ class Xhotkeys:
 		self.widgets["accept"].set_sensitive(sensitive)
 
 	#######################################
-	def open_add_window(self):
-		self.widgets["mainwindow"].set_sensitive(False)
+	def open_add_window(self, modify=False, position = None):
+		self.add_position = position
+		try: self.active_row = self.get_selected_rows()[1][0][0]
+		except: self.active_row = None
 		self.widgets["name"].set_text(self.current_comb["name"])
+		if modify: self.new_window = "modify"; self.widgets["change"].grab_focus()
+		else: self.new_window = "add"; self.widgets["name"].grab_focus()
 		self.widgets["command"].set_text(self.current_comb["command"])
 		self.set_hotkey_text(self.current_comb["hotkey"])
 		self.widgets["addwindow"].set_transient_for(self.widgets["mainwindow"])
@@ -506,11 +486,11 @@ class Xhotkeys:
 		self.debug("creating pidfile: %s" %self.pidfile)
 		try: fd = open(self.pidfile, "w")
 		except: self.debug("cannot open file for writing: %s" %self.pidfile, ERROR, exit =1)
-		pid = fd.write(str(os.getpid()) + "\n")
+		fd.write(str(os.getpid()) + "\n")
 		fd.close()
 		
 	#######################################
-	def check_active(self, stop=False):
+	def check_active(self):
 		try: fd = open(self.pidfile)
 		except IOError: self.debug("no pidfile found"); return
 		try: pid = int(fd.readline().strip())
@@ -530,17 +510,14 @@ class Xhotkeys:
 
 	#######################################
 	def stop_server(self):
-		self.debug("stopping xhotkey server")
+		self.debug("closing xhotkey server")
 		self.hkserver.stop()
-		self.hkserver.join()
-		self.debug("hotkey server finished")
 
 	#######################################
 	def exit_conf(self):
 		self.save_settings()
+		self.write_configuration()
 		gtk.main_quit()
-		if self.hkserver and self.hkserver.isAlive():
-			self.stop_server()
 		os._exit(0)
 
 	#######################################		
@@ -555,8 +532,8 @@ class Xhotkeys:
 			getattr(self.widgets[key], function)(value)
 
 	#########################################
-	def get_current_rows(self):
-		return self.treeselection.get_selected_rows()[1]
+	def get_selected_rows(self):
+		return self.widgets["hotkeytable"].get_selection().get_selected_rows()
 	
 	#######################################
 	def delete_entry(self, widget, answer, iters, rows):
@@ -569,37 +546,19 @@ class Xhotkeys:
 					self.combinations.remove(entry)
 			self.apply_widgets("set_sensitive", delete=False, modify=False)
 			self.save_and_reload()
-		self.widgets["mainwindow"].set_sensitive(True)
 
 	#######################################
-	def drag_received(self, widget, context, x, y, selection, targetType, time):
-		path = urllib.url2pathname(selection.data)
-		filestr = ('file:\\\\\\', 'file://', 'file:')
-		for s in filestr:
-			if path.startswith(s):
-				path = path[len(s):].replace('\n', '').replace('\r', '')
-				break
-		if not path.endswith(".desktop"): return
-		try: fd = open(path)
-		except: self.debug("error opening desktop file: %s" %path, INFO); return
-		namelist, lang = ["Name"], os.getenv("LANG")
-		if lang: namelist.append("Name[%s]" %lang)
-		if lang.find("_") > 0: namelist.append("Name[%s]" %(lang[:lang.find("_")]))
-		names, command = [], ""
-		for line in fd:
-			line = line.strip()
-			if not line: continue
-			try: key, value = line.split("=")
-			except: continue
-			if key in namelist:
-				names.append([key, value.strip()])
-			elif not command and key == "Exec":
-				command = value.strip()
-		fd.close()
-		if not command or not names: self.debug("error parsing desktop file: %s" %path, INFO); return
-		names.sort()
-		name = names[-1][1]
-		return name, command
+	def hide_order(self):
+		for col in self.widgets["hotkeytable"].get_columns():
+			col.set_sort_indicator(False)
+			self.column_order = None
+
+	#########################################
+	def hotkeytable_selection_changed(self, *args):
+		rows = self.get_selected_rows()[1]
+		if len(rows) == 0: self.apply_widgets("set_sensitive", delete=False, modify=False)
+		elif len(rows) == 1: self.apply_widgets("set_sensitive", delete=True, modify=True)
+		else: self.apply_widgets("set_sensitive", delete=True, modify=False)
 
 	#######################################
 	## WIDGETS ###############################
@@ -620,19 +579,58 @@ class Xhotkeys:
 		self.exit_conf()
 
 	#######################################
-	def on_hotkeytable_drag_data_received(self, *args):
-		ret = self.drag_received(*args)
-		if not ret: return
-		name, command = ret
+	def on_hotkeytable_drag_data_received(self, treeview, context, x, y, selection, targetType, etime):
+		target = selection.target
+		model = treeview.get_model()
+		drop_info = treeview.get_dest_row_at_pos(x, y)
+		if drop_info: path, position = drop_info; dest_row = path[0]
+		else: dest_row = len(self.combinations)
+		if target == "xhotkeys/row" and self.rows_to_move:
+			if drop_info and position not in (gtk.TREE_VIEW_DROP_BEFORE, 
+					gtk.TREE_VIEW_DROP_INTO_OR_BEFORE, gtk.TREE_VIEW_DROP_INTO_OR_AFTER ):
+				dest_row += 1
+			combs1 = [self.combinations[x] for x in range(dest_row) if x not in self.rows_to_move]
+			combs2 = [self.combinations[x] for x in self.rows_to_move]
+			combs3 = [self.combinations[x] for x in range(len(self.combinations)) if x >= dest_row and x not in self.rows_to_move]
+			newcombs = combs1 + combs2 + combs3
+			if self.combinations == newcombs: return
+			self.combinations = newcombs
+			self.fill_table(keep_order=False)
+			if context.action == gtk.gdk.ACTION_MOVE:
+				context.finish(True, True, etime)
+			return
+		if target != "text/uri-list": return
+		path = urllib.url2pathname(selection.data)
+		filestr = ('file:\\\\\\', 'file://', 'file:')
+		try: path = [path[len(s):].replace('\n', '').replace('\r', '') for s in filestr if path.startswith(s)][0]
+		except: self.debug("not a uri-file: %s" %path.strip()); return
+		if not path.endswith(".desktop"): self.debug("not a desktop file: %s" %path.strip()); return
+		try: fd = open(path)
+		except: self.debug("error opening desktop file: %s" %path.strip(), INFO); return
+		namelist, lang = ["Name"], os.getenv("LANG")
+		if lang: namelist.append("Name[%s]" %lang)
+		if lang.find("_") > 0: namelist.append("Name[%s]" %(lang[:lang.find("_")]))
+		names, command = [], ""
+		for line in fd:
+			line = line.strip()
+			try: key, value = line.split("=")
+			except: continue
+			if key in namelist:
+				names.append([key, value.strip()])
+			elif not command and key == "Exec":
+				command = value.strip()
+		fd.close()
+		if not command or not names: 
+			self.debug("error parsing desktop file: %s" %path, INFO); return
+		names.sort()
+		name = names[-1][1]
 		hotkey = self.keycodes_to_string([])
 		self.current_comb = create_dict(name=name, command=command, hotkey=hotkey)
-		self.old_comb = None
-		self.open_add_window()
+		self.open_add_window(position=dest_row)
 		self.old_comb = None
 		self.textbox_check(self.widgets["name"], "name")
 		self.textbox_check(self.widgets["command"], "command")
-		self.new_window = "add"
-
+			
 	#######################################
 	def on_accept_clicked(self, widget):
 		for index, combination in enumerate(self.combinations):
@@ -640,22 +638,19 @@ class Xhotkeys:
 			if combination["name"] == self.current_comb["name"]:
 				self.error_message("This hotkey name (%s) is already being used" %combination["name"])
 				return
-
 		if self.new_window == "modify":
 			self.combinations[self.active_row] = self.current_comb
-			rows = self.get_current_rows()
-			treemodel, selected = self.treeselection.get_selected_rows()
+			treemodel, selected = self.widgets["hotkeytable"].get_selection().get_selected_rows()
 			iter = [treemodel.get_iter(path) for path in selected][0]
-			treemodel.set_value(iter, 0, self.current_comb["name"])
-			treemodel.set_value(iter, 1, self.current_comb["command"])
-			treemodel.set_value(iter, 2, self.current_comb["hotkey"].title())
+			self.fill_table()
 		else:
-			self.combinations.append(self.current_comb)
-			self.liststore.insert(len(self.combinations)-1, [self.current_comb["name"], self.current_comb["command"], self.current_comb["hotkey"].title()])
-			self.treeview.set_cursor(len(self.combinations)-1)
+			if self.add_position != None: position = self.add_position
+			else: position = len(self.combinations)
+			self.combinations.insert(position, self.current_comb)
+			self.fill_table()
+			self.widgets["hotkeytable"].set_cursor(position)
 
 		self.widgets["addwindow"].hide()
-		self.widgets["mainwindow"].set_sensitive(True)
 		
 		if self.old_comb != self.current_comb:
 			self.save_and_reload()
@@ -663,19 +658,16 @@ class Xhotkeys:
 	#######################################
 	def on_cancel_clicked(self, widget):
 		self.widgets["addwindow"].hide()
-		self.widgets["mainwindow"].set_sensitive(True)
 		if self.old_comb:
 			self.current_comb = self.old_comb.copy()
 
 	#######################################
 	def on_modify_clicked(self, widget):
-		row = self.get_current_rows()[0][0]
-		self.active_row = row
+		row = self.get_selected_rows()[1][0][0]
 		self.current_comb = self.combinations[row].copy()
 		self.old_comb = self.current_comb.copy()
-		treemodel, selected = self.treeselection.get_selected_rows()
-		self.open_add_window()
-		self.new_window = "modify"
+		treemodel, selected = self.get_selected_rows()
+		self.open_add_window(modify=True)
 
 	#######################################
 	def on_test_clicked(self, widget):
@@ -687,10 +679,8 @@ class Xhotkeys:
 		hotkey = self.keycodes_to_string([])
 		self.current_comb = create_dict(name="", command="", hotkey=hotkey)
 		self.apply_widgets("set_sensitive", accept=False, test=False)		
-		self.widgets["name"].grab_focus()
-		self.old_comb = None
 		self.open_add_window()
-		self.new_window = "add"
+		self.old_comb = None
 		
 	#######################################
 	def on_change_clicked(self, widget):
@@ -740,7 +730,7 @@ class Xhotkeys:
 		widget.stop_emission("key_press_event")
 
 	#########################################
-	def on_status_key_release_event(self, widget, event):
+	def on_status_delete_event(self, widget, event):
 		pass
 
 	#########################################
@@ -753,10 +743,9 @@ class Xhotkeys:
 
 	#######################################
 	def on_delete_clicked(self, widget):
-		rows = self.get_current_rows()
-		model, selected = self.treeselection.get_selected_rows()
-		iters = [model.get_iter(path) for path in selected]
-		names = [self.combinations[row[0]]["name"] for row in rows]
+		model, rows = self.get_selected_rows()
+		iters = [model.get_iter(path) for path in rows]
+		names = [" * "+self.combinations[row[0]]["name"] for row in rows]
 		todelete = "\n".join(names)
 		message = gtk.MessageDialog(self.widgets["mainwindow"], gtk.DIALOG_MODAL, \
 			gtk.MESSAGE_WARNING, gtk.BUTTONS_YES_NO, "Are you sure you want to delete the following entries?\n%s" \
@@ -764,7 +753,6 @@ class Xhotkeys:
 		message.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
 		rowstodelete = [row[0] for row in rows]
 		message.connect("response", self.delete_entry, iters, rowstodelete)
-		self.widgets["mainwindow"].set_sensitive(False)
 		message.show()
 
 	#######################################
@@ -787,38 +775,38 @@ class Xhotkeys:
 		self.on_modify_clicked(self.widgets["modify"])
 
 	#######################################
-	def on_hotkeytable_button_press_event(self, treeview, event):
-		if not treeview.get_path_at_pos(int(event.x), int(event.y)): 
-			if event.button != 3: self.treeselection.unselect_all()
-			else: items = ("gtk-add", self.on_add_clicked),
-		elif event.button == 3: 
-			self.rbutton_event = event.time
-	
-	#########################################
-	def hotkeytable_selection_changed(self, *args):
-		rows = self.get_current_rows()
-		if self.rbutton_event:
-			menu = gtk.Menu()
+	def on_hotkeytable_button_press_event(self, widget, event):
+		rows = [x[0] for x in self.get_selected_rows()[1]]
+		object_path = widget.get_path_at_pos(int(event.x), int(event.y))
+		if not object_path and event.button == 1: 
+			widget.get_selection().unselect_all()
+			return
+		if event.button != 3: return
+		in_selection = True
+		if object_path and event.button == 3: 
+			mouse_path = object_path[0][0]
+			if mouse_path not in rows: 
+				in_selection = False
+				rows = [mouse_path]
 			items = [("gtk-delete", self.on_delete_clicked)]
-			if len(rows) == 1:
-				items.append(("gtk-edit", self.on_modify_clicked))
-			for stock_id, connect in items:
-				item = gtk.ImageMenuItem(stock_id=stock_id)
-				menu.append(item)
-				item.connect("activate", connect)
-				item.show()
-			menu.popup(None, None, None, 3, self.rbutton_event)
-			self.rbutton_event = None
-		if len(rows) == 0: self.apply_widgets("set_sensitive", delete=False, modify=False)
-		elif len(rows) == 1: self.apply_widgets("set_sensitive", delete=True, modify=True)
-		else: self.apply_widgets("set_sensitive", delete=True, modify=False)
+			if len(rows) == 1: items.append(("gtk-edit", self.on_modify_clicked))
+		else: items = [("gtk-add", self.on_add_clicked)]
+		menu = gtk.Menu()
+		for stock_id, connect in items:
+			item = gtk.ImageMenuItem(stock_id=stock_id)
+			menu.append(item)
+			item.connect("activate", connect)
+			item.show()
+		menu.popup(None, None, None, 3, event.time)
+		if in_selection and rows: return True
+	
 	
 	#####################################################
 	#####################################################
 
 	#############################################
 	def configure_signals(self, signals):
-		# Create signals dictionary: key = <signal_int> - value = <signal_name">
+		"""Create signals dictionary: {signal_int: signal_name}"""
 		for signum in signals:
 			signal.signal(signum, self.signal_handler)
 		self.signames = {}
@@ -840,11 +828,11 @@ class Xhotkeys:
 
 	#####################################################
 	def open_glade(self, files):
-		for file in files:
-			try: os.path.exists(file)
-			except: self.debug("not found: %s" %file); continue
-			try: self.tree = gtk.glade.XML(file)
-			except: self.debug("cannot load xml: %s" %file, ERROR); continue
+		for path in files:
+			try: os.path.exists(path)
+			except: self.debug("not found: %s" %path); continue
+			try: self.tree = gtk.glade.XML(path)
+			except: self.debug("cannot load xml: %s" %path, ERROR); continue
 			else: break
 		else: self.debug("no glade files found", ERROR, exit=1)
 
@@ -857,17 +845,45 @@ class Xhotkeys:
 			if not self.widgets.has_key(widget):
 				self.widgets[widget] = self.tree.get_widget(widget)
 			self.widgets[widget].connect(sig, getattr(self, item))
+
+	#####################################################
+	def column_clicked(self, widget):
+		column_name  = widget.get_title().lower()
+		if widget.get_sort_indicator() and widget.get_sort_order() == gtk.SORT_ASCENDING: 
+			order = gtk.SORT_DESCENDING
+		else:order = gtk.SORT_ASCENDING
+		self.column_order = column_name, order
+		self.fill_table()
+
+	#####################################################
+	def order_combinations(self):
+		if not self.column_order: return
+		column_name, order = self.column_order
+		index = [x[0].lower() for x in TABLE].index(column_name)
+		columns = self.widgets["hotkeytable"].get_columns()
+		widget = columns[index]
+		widget.set_sort_order(order)
+		for col in columns:
+			col.set_sort_indicator(col.get_title().lower() == column_name)
+		column_values = [(x[column_name].lower(), x) for x in self.combinations]
+		column_values.sort()
+		if order == gtk.SORT_DESCENDING: column_values.reverse()
+		self.combinations = [x[1] for x in column_values]
 		
+	#######################################
+	def on_hotkeytable_drag_begin(self, widget, context):
+		self.rows_to_move = [x[0] for x in self.get_selected_rows()[1]]
+
 	#####################################################
 	def create_table(self, widget_name, options):
-		self.treeview = self.tree.get_widget(widget_name)
-		self.treeselection = self.treeview.get_selection()
-		self.rbutton_event =None
-		self.treeselection.connect("changed", self.hotkeytable_selection_changed)
-		self.treeselection.set_mode(gtk.SELECTION_MULTIPLE)
+		self.column_order = None
+		treeview = self.tree.get_widget(widget_name)
+		treeselection = treeview.get_selection()
+		treeselection.connect("changed", self.hotkeytable_selection_changed)
+		treeselection.set_mode(gtk.SELECTION_MULTIPLE)
 		format = tuple([str]) * len(options)
 		self.liststore = gtk.ListStore(*format)
-		self.treeview.set_model(self.liststore)			
+		treeview.set_model(self.liststore)
 		for index, (name, min_width, expand) in enumerate(options):
 			renderer = gtk.CellRendererText()
 			column = gtk.TreeViewColumn(name, renderer, text=index)
@@ -876,23 +892,27 @@ class Xhotkeys:
 			column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
 			column.set_expand(expand)
 			column.set_min_width(min_width)
-			self.treeview.append_column(column)
+			column.connect("clicked", self.column_clicked)
+			treeview.append_column(column)
 		self.apply_widgets("set_sensitive", delete=False, modify=False)
-		list = [ ( 'text/uri-list', 0, TARGET_TYPE_URI_LIST ) ]
-		self.treeview.drag_dest_set(gtk.DEST_DEFAULT_MOTION | gtk.DEST_DEFAULT_HIGHLIGHT | \
-			gtk.DEST_DEFAULT_DROP, list, gtk.gdk.ACTION_COPY)
-
+		
+		# Enable dragging desktop files to table
+		mime_list = [ ( 'text/uri-list', 0, TARGET_TYPE_URI_LIST ), ('xhotkeys/row', gtk.TARGET_SAME_WIDGET, 0) ]
+		treeview.enable_model_drag_source(gtk.gdk.BUTTON1_MASK, \
+			[('xhotkeys/row', gtk.TARGET_SAME_WIDGET, 0)], gtk.gdk.ACTION_COPY)
+		treeview.enable_model_drag_dest(mime_list, gtk.gdk.ACTION_COPY)
 
 	#####################################################
 	def save_settings(self):
 		try: fd = open(self.settingsfile, "w")
 		except: self.debug("settings could not be opened for writing: %s" %self.settingsfile); return
-		fd.write("# automatically saved settings xhotkeys\n")
+		fd.write("# xhotkeys: automatically saved settings\n")
 		x, y = self.widgets["mainwindow"].get_position()
 		width, height = self.widgets["mainwindow"].get_size()
 		fd.write(SETTINGS_WINDOW % (x, y, width, height) + "\n")
 		widths = tuple([column.get_width() for column in self.widgets["hotkeytable"].get_columns()])
 		fd.write(SETTINGS_COLUMN % widths + "\n")
+		if self.column_order: fd.write(SETTINGS_ORDER % self.column_order + "\n")
 		fd.close()
 		self.debug("window settings saved")
 				
@@ -900,8 +920,10 @@ class Xhotkeys:
 	def load_settings(self):
 		try: fd = open(self.settingsfile)
 		except: self.debug("settings file not found: %s" %self.settingsfile); return
-		match_window = SETTINGS_WINDOW.replace(" ", "\s+").replace("(", "\(").replace(")", "\)").replace("%d", "(\d+)")
-		match_column = SETTINGS_COLUMN.replace(" ", "\s+").replace("%d", "(\d+)")
+		match_window = SETTINGS_WINDOW.replace("(", "\(").replace(")", "\)").replace("%d", "(\d+)")
+		match_column = SETTINGS_COLUMN.replace("%d", "(\d+)").replace("%d", "(\S+)")
+		match_order = SETTINGS_ORDER.replace("%d", "(\d+)").replace("%s", "(\w+)")
+		columns = self.widgets["hotkeytable"].get_columns()
 		for line in fd.readlines():
 			match = re.findall(match_window, line)
 			if match:
@@ -913,26 +935,39 @@ class Xhotkeys:
 			if match:
 				try: widths = [int(x) for x in match[0]]
 				except: continue
-				columns = self.widgets["hotkeytable"].get_columns()
 				[columns[x].set_fixed_width(widths[x]) for x in range(len(columns))]
+			match = re.findall(match_order, line)
+			if match and len(match[0]) == 2:
+				try: name, order = match[0]
+				except: continue
+				self.column_order = name, int(order)
+				try: index = [x[0].lower() for x in TABLE].index(name)
+				except: continue
+				columns[index].set_sort_order(int(order))
+				columns[index].set_sort_indicator(True)
 		fd.close()
 		self.debug("window settings loaded")
-			
+
+	#####################################################
+	def fill_table(self, keep_order=True):
+		if keep_order: self.order_combinations()
+		else: self.hide_order()
+		self.liststore.clear()
+		for comb in self.combinations:
+			self.liststore.append([comb["name"], comb["command"], comb["hotkey"].title()])
+
 	#####################################################
 	def config(self):
 		self.state = "config"
 		self.daemon_pid = self.check_active()	
-		glade_files = [NAME + ".glade", os.path.join("/usr/share/", NAME, NAME + ".glade")]
+		glade_files = [NAME + ".glade", os.path.join("/usr/share", NAME, NAME + ".glade")]
 		self.open_glade(glade_files)
 		self.load_widgets()
-		table = ("Name", 100, False), ("Command", 300, True), ("Hotkey", 150, True)
 		self.read_configuration(self.cfile)
-		self.create_table("hotkeytable", table)		
+		self.create_table("hotkeytable", TABLE)		
 		self.load_settings()
-		for comb in self.combinations:
-			self.liststore.append([comb["name"], comb["command"], comb["hotkey"].title()])
-		if self.check_active():
-			self.widgets["status"].push(0, " Xhotkeys daemon is running (pid %d)" %self.daemon_pid)
+		self.fill_table()
+		if self.check_active(): self.widgets["status"].push(0, " Xhotkeys daemon is running (pid %d)" %self.daemon_pid)
 		else: self.widgets["status"].push(0, " Xhotkeys daemon is not running")
 		
 		self.configure_signals([signal.SIGTERM, signal.SIGINT])
@@ -944,25 +979,23 @@ class Xhotkeys:
 		self.state = "server"
 		if not os.path.exists(self.cfile):
 			self.debug("configuration file not found: %s" %self.cfile, ERROR)
-			self.debug("run %s --config" %NAME, ERROR, 1)
-		else: self.debug("using configuration file: %s" %self.cfile, INFO)
+			self.debug("to configure it run %s --config" %NAME, ERROR)
 		self.hkserver = None
 		pid = self.check_active()
 		if pid: 
 			self.debug("oops, %s daemon is already running (pid %s)" %(NAME, pid), ERROR)
 			self.debug("run '%s --config' to open configurator" %NAME, ERROR, exit=1)
-			
-		self.debug("starting Xhotkeys server")
+		self.debug("Starting Xhotkeys server")
 		self.configure_signals([signal.SIGTERM, signal.SIGINT, signal.SIGHUP])
 		self.hkserver = XhotkeysServer(self.verbose_level)
 		self.read_configuration(self.cfile)
 		self.hkserver.set_combination(self.combinations)
-		self.hkserver.start()
 		self.create_pidfile()
 		while 1:
-			try: self.hkserver.join(0.1)
+			try: self.hkserver.run()
 			except: break
-			
+			# Give the CPU a breath (1/4 second should be fine)
+			time.sleep(0.25)
 		self.delete_pidfile()
 		os._exit(0)
 
